;; Tin Can Receiver
;;
;; Copyright 2020 - By Michael Kohn
;; http://www.mikekohn.net/
;; mike@mikekohn.net
;;
;; A microphone into an LM567 will listen for tones at 400Hz
;; with encoded data and transmit to a computer over UART.

.include "msp430x2xx.inc"

RAM equ 0x0200
UART_CALIBRATE equ RAM
UART_CALIBRATE_RAW equ RAM+2

DATA equ 0x01
DEBUG_LED equ 0x08

; 2500 cycles between interrupts
; 0.00125 seconds between interrupts.
; Start = 42 interrupts = 0.052 seconds
; Start = 28 interrupts = 0.035 seconds
; Start = 21 interrupts = 0.02625 seconds
; Start = 14 interrupts = 0.0175 seconds
;BIT_START equ 42
;BIT_ONE   equ 28
;BIT_ZERO  equ 21
;BIT_SPACE equ 14

; Start = 160 interrupts = 0.200 seconds
;   One =  80 interrupts = 0.100 seconds
;  Zero =  60 interrupts = 0.075 seconds
; Space =  40 interrupts = 0.050 seconds

BIT_START equ 160
BIT_ONE   equ 80
BIT_ZERO  equ 60
BIT_SPACE equ 40

THRESHOLD_UPPER equ 20
THRESHOLD_LOWER equ 5

;  r4 = detect count (updated in interrupt)
;  r5 =
;  r6 =
;  r7 =
;  r8 = data bit counter
;  r9 = count interrupt_b
; r10 =
; r11 = count in interrupt_a
; r12 = data byte
; r13 =
; r14 =
; r15 = temp

.org 0xc000
start:
  ;; Turn off watchdog
  mov.w #WDTPW|WDTHOLD, &WDTCTL

  ;; Turn off interrupts
  dint

  ;; Setup stack pointer
  mov.w #0x0400, SP

  ;; Set MCLK to 2 MHz with DCO
  mov.b #DCO_3, &DCOCTL
  mov.b #RSEL_9, &BCSCTL1
  mov.b #0, &BCSCTL2

  ;; Set SMCLK to 32.768kHz external crystal
  mov.b #XCAP_3, &BCSCTL3

  ;; Setup output pins
  ;; P1.0 = DATA (from LM567)
  ;; P1.1 = RX
  ;; P1.2 = TX
  ;; P2.2 = DEBUG_LED
  ;; P2.6 = 32.768Hz Crystal
  ;; P2.7 = 32.768Hz Crystal
  mov.b #0, &P1DIR
  mov.b #0, &P1OUT
  mov.b #0x06, &P1SEL
  mov.b #0x06, &P1SEL2
  mov.b #DEBUG_LED, &P2DIR
  mov.b #DEBUG_LED, &P2OUT

  ;; Setup UART (9600 @ 2 MHz)
  mov.b #UCSSEL_2|UCSWRST, &UCA0CTL1
  mov.b #0, &UCA0CTL0
  mov.b #UCBRS_2, &UCA0MCTL
  mov.b #220, &UCA0BR0
  mov.b #0, &UCA0BR1
  bic.b #UCSWRST, &UCA0CTL1

  ;; Setup Timer A
  mov.w #0xffff, &TACCR0
  mov.w #TASSEL_2|MC_1, &TACTL    ; TACLK, DIV1, COUNT to 0xffff
  ;mov.w #CCIE, &TACCTL0
  mov.w #0, &TACCTL0

  ;; Setup Timer B
  mov.w #436, &TBCCR0          ; 32768 ticks = 1 second
  mov.w #TBSSEL_1|MC_1, &TBCTL ; ACLK, DIV1, COUNT to TBCCR0
  ;mov.w #CCIE, &TBCCTL0
  mov.w #0, &TBCCTL0

  ;; Turn on interrupts
  eint

  ;; To help debugging, clear out some RAM.
  mov.w #0x200, r15
memset:
  mov.w #0, 0(r15)
  add.w #2, r15
  cmp.w #0x230, r15
  jnz memset

  ;; Run twice.. DCO isn't stable?
  call #calibrate_uart
  call #calibrate_uart

  ;; Stop interrupt on 32.768kHz crystal.
  bic.w #CCIE, &TBCCTL0

  ;; Interrupt every 0.0125 seconds to match transmitter.
  dint
  bis.w #TACLR, &TACTL
  mov.w #0, &TACTL
  mov.w #2500, &TACCR0
  ;mov.w #2608, &TACCR0
  mov.w #TASSEL_2|MC_1, &TACTL    ; TACLK, DIV1
  mov.w #CCIE, &TACCTL0
  ;mov.w #0, &TACCTL0
  eint

  ;; DEBUG
  mov.w #0, r6
  mov.w #0, r4
  mov.w #0xffff, r8

  mov.b #'A', r15
  call #uart_send_char

  mov.w #RAM+16, r10

main:
  cmp.w #THRESHOLD_UPPER, r4
  jnz main
  call #read_data
  jmp main

read_data:
  bic.b #DEBUG_LED, &P2OUT
  mov.w #RAM+14, r10

mov.w #1, r5 ;; DEBUG
  ;; Wait for start bit.
  mov.w #0, r11
read_data_wait_start:
  cmp.w #BIT_START+25, r11
  jhs read_data_exit
  cmp.w #THRESHOLD_LOWER, r4
  jhs read_data_wait_start
mov.w #2, r5 ;; DEBUG
  cmp.w #BIT_START-50, r11
  jlo read_data_exit

  mov.w r11, 0(r10)
  add.w #2, r10

  ;; Wait on a space.
  ;; If start bit is too short, exit.
  ;cmp.w #BIT_START-25, r11
  ;jlo read_data_exit

mov.w #3, r5 ;; DEBUG

  ;; Read 8 bits.
  mov.w #0, r15
  mov.w #0, r12
  mov.w #8, r8
read_data_loop:

mov.w #12, r5 ;; DEBUG

  ;; Wait on a space.
  mov.w #0, r11
read_data_wait_space:
  cmp.w #BIT_ONE+5, r11
  jhs read_data_exit
  cmp.w #THRESHOLD_UPPER, r4
  jlo read_data_wait_space
  ;; If space is too short, exit.
  ;cmp.w #BIT_SPACE-5, r11
  ;jlo read_data_exit

mov.w #10, r5 ;; DEBUG
  ;; Read a data bit.
  mov.w #0, r11
read_data_wait_bit:
  cmp.w #BIT_ONE+25, r11
  jhs read_data_exit
  cmp.w #THRESHOLD_LOWER, r4
  jhs read_data_wait_bit

mov.w #32, r5 ;; DEBUG

  ;; If bit is too short, exit.
  cmp.w #BIT_ZERO-30, r11
  jlo read_data_exit

  ;; Shift data byte.
  rla.w r12
  ;; If bit is shorter than a one, don't set bit 0.
  ;cmp.w #BIT_ZERO-20, r11
  cmp.w #0x2f, r11
  jlo read_data_not_1
  bis.w #1, r12
read_data_not_1:

  ;; DEBUG
  mov.w r11, 0(r10)
  add.w #2, r10
  ;; DEBUG

  ;; If space is too short, exit.
  ;cmp.w #BIT_SPACE-5, r11
  ;jlo read_data_exit
  ;; Next bit.
  dec.w r8
  jnz read_data_loop

  mov.b r12, r15
  call #uart_send_char
read_data_exit:
  bis.b #DEBUG_LED, &P2OUT
mov.w r11, r6 ;; DEBUG
  ret

; uart_send_char(r15)
uart_send_char:
  bit.b #UCA0TXIFG, &IFG2
  jz uart_send_char
  mov.b r15, &UCA0TXBUF
  ret

;;    9600 / 128 = 75, 128 / 9600 = 0.0039062
;;   32768 / 128 = 256, (128 / 9600 ) * 32768 = 436.91
;; 2000000 / 128 = 15625
;; 2000000 / 9600 baud = 208.33 (looking for ~208 ticks on TAR)
;; Set 32678 timer to count to 256, after 256 count ticks on 2MHz timer.
;; Divide the count on 2MHz by 75 to get count of ticks for 1 bit on 9600 baud.
calibrate_uart:
  bis.w #TBCLR, &TBCTL
  bis.w #CCIE, &TBCCTL0
  bis.w #TACLR, &TACTL
  ;bic.w #CCIE, &TACCTL0
  mov.w #0, r9
calibrate_uart_loop:
  mov.w &TAR, r14
  cmp.w #1, r9
  jnz calibrate_uart_loop
  mov.w r14, &UART_CALIBRATE_RAW
  mov.w #0, r15
calibrate_uart_divide_75:
  cmp.w #128, r14
  jlo calibrate_uart_divide_75_done
  sub.w #128, r14
  add.w #1, r15
  jmp calibrate_uart_divide_75
calibrate_uart_divide_75_done:
  mov.w r15, &UART_CALIBRATE
  mov.w r15, &UCA0BR0
  ret

timer_interrupt_a:
  inc.w r11
  bit.b #1, &P1IN
  jnz timer_interrupt_a_sub_1
  cmp.w #THRESHOLD_UPPER, r4
  jz timer_interrupt_a_exit
  add.w #1, r4
  reti
timer_interrupt_a_sub_1:
  cmp.w #0, r4
  jz timer_interrupt_a_exit
  sub.w #1, r4
timer_interrupt_a_exit:
  reti

timer_interrupt_b:
  inc.w r9
  reti

.org 0xfff2
  dw timer_interrupt_a     ; Timer0_A3 TACCR0, CCIFG
.org 0xfffa
  dw timer_interrupt_b     ; Timer1_A3 TACCR0, CCIFG
.org 0xfffe
  dw start                 ; Reset

